# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import os
import io
from pathlib import Path
import json

from flask import Flask, request, jsonify, send_file, send_from_directory
from flask_cors import CORS
from ariadne.explorer import ExplorerGraphiQL
from ariadne import combine_multipart_data, graphql_sync

import torch
from torchvision.transforms import ToPILImage

from equine_webapp.utils import SERVER_CONFIG, get_support_example_from_data_index, get_sample_from_data_index, sanitize_path
from equine_webapp.graphql.graphql_config import schema

# Flask App Setup ################################
app = Flask(__name__)
# app.config["CORS_HEADERS"] = "Content-Type"
CORS(app)

# GraphQL Setup ##################################
explorer_html = ExplorerGraphiQL().html(None)


# Helper Functions ###############################
def start_dev_server():
    debug_mode = os.getenv('FLASK_DEBUG', 'False').lower() in ['true', '1', 't']
    app.run(port=8080, debug=debug_mode)


# App Routes #####################################
@app.route("/graphql", methods=["GET"])
def graphql_explorer():
    # On GET request serve the GraphQL explorer.
    return explorer_html, 200

@app.route("/graphql", methods=["POST"])
def graphql_server():
    if request.content_type.startswith("multipart/form-data"):
        data = combine_multipart_data(
            json.loads(request.form.get("operations")),
            json.loads(request.form.get("map")),
            dict(request.files)
        )
    else:
        data = request.get_json()
    success, result = graphql_sync(
        schema,
        data,
        context_value=request,
        debug=app.debug
    )
    status_code = 200 if success else 400
    return jsonify(result), status_code

@app.route("/api/render-image/inference/<run_id>/<data_index>", methods=["GET"])
def handle_render_inference_image(run_id, data_index):
    sample, _, _ = get_sample_from_data_index(run_id, data_index)
    return send_img_tensor_as_file(sample, f"{run_id}_{data_index}")

    

@app.route("/api/render-image/support/<model_name>/<data_index>", methods=["GET"])
def handle_render_support_image(model_name, data_index):
    support_example, _, _ = get_support_example_from_data_index(model_name, data_index)
    return send_img_tensor_as_file(support_example, f"{model_name}_{data_index}")

def send_img_tensor_as_file(img_tensor, filename):
    img = ToPILImage()(img_tensor.to(torch.uint8))
    img_buf = io.BytesIO()
    img.save(img_buf, format="png")
    img_buf.seek(0)

    return send_file(img_buf, download_name=f"{filename}.png")

@app.route("/api/models/<model_name>", methods=["GET"])
def handle_send_model(model_name):
    return send_file(os.path.join(os.getcwd(), SERVER_CONFIG.MODEL_FOLDER_PATH, model_name), download_name=model_name)


# HTML / static routing #####################################
@app.route('/')
def index():
    return send_from_directory('client', 'index.html')

@app.route('/<path:path>')
def route_client(path):
    path = sanitize_path(path, "")
    root, ext = os.path.splitext(path)
    if not ext:
        path += ".html"
    return send_from_directory('client', path)



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import random
from equine_webapp.tests.train_model_for_testing import TEST_MODEL_CONFIG
from equine_webapp.tests.utils import assert_confidence_labels_are_valid

def test_query_models(client):
    # create a NUM_VECTORS x VECTOR_SIZE matrix of random floats
    NUM_VECTORS = 30
    VECTOR_SIZE = 10
    data = []
    for i in range(NUM_VECTORS):
        data.append([random.uniform(-100, 100) for _ in range(VECTOR_SIZE)])

    response = client.post("/graphql", json={
        "query": """
          query DimensionalityReduction(
            $method: String!, 
            $data:[[Float!]!]!, 
            $nNeighbors: Int!,
          ) {
            dimensionalityReduction(method: $method, data:$data, nNeighbors:$nNeighbors) {
              continuity
              embeddings
              stress
              scree
              srho
              trustworthiness
            }
          }
        """,
        "variables": {
            "method": "pca",
            "data": data,
            "nNeighbors": 5,
        },
    })

    response_data = response.json["data"]["dimensionalityReduction"]

    assert response_data["continuity"] >= 0.0
    assert response_data["continuity"] <= 1.0

    assert response_data["stress"] >= 0.0

    assert response_data["srho"] >= 0.0
    assert response_data["srho"] <= 1.0

    assert response_data["trustworthiness"] >= 0.0
    assert response_data["trustworthiness"] <= 1.0

    assert len(response_data["embeddings"]) == NUM_VECTORS
    for embedding_2d in response_data["embeddings"]:
        assert len(embedding_2d) == 2 # correct embedding vector size
        all(isinstance(item, float) for item in embedding_2d) # all coordinates are floats

    assert len(response_data["scree"]) == 5
    all(isinstance(item, float) for item in response_data["scree"])
    all(item >= 0 for item in response_data["scree"])
    all(item <= 1 for item in response_data["scree"])


# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
from zadu import zadu


# dimensionality reduction
import numpy as np
from sklearn.manifold import MDS, TSNE
from sklearn.decomposition import PCA
import umap


import equine as eq
import torch
import os
from ariadne import convert_kwargs_to_snake_case

from equine_webapp.utils import SERVER_CONFIG, get_support_example_from_data_index, get_sample_from_data_index, get_model_path, use_label_names

@convert_kwargs_to_snake_case
def resolve_available_models(_, info, extension):
    model_folder = os.path.join(os.getcwd(), SERVER_CONFIG.MODEL_FOLDER_PATH)

    filter_extension = extension if extension else SERVER_CONFIG.MODEL_EXT
    model_files = sorted([x for x in os.listdir(model_folder) if filter_extension in x])

    request_data = []
    for file_name in model_files:
        model_path = os.path.join(model_folder, file_name)
        last_mod_time = os.path.getmtime(model_path)
        request_data.append({"name": file_name, "lastModified" : last_mod_time})

    return request_data

@convert_kwargs_to_snake_case
def resolve_model_summary(_, info, model_name):
    model_path = get_model_path(model_name)
    model_save = torch.load(model_path)
    summary = model_save["train_summary"]
    summary["lastModified"] = os.path.getmtime(model_path)
    return summary

@convert_kwargs_to_snake_case
def resolve_get_protonet_support_embeddings(_, info, model_name):
    model_path = get_model_path(model_name)
    model = eq.load_equine_model(model_path)
    support_examples = model.get_support()
    prototypes = model.get_prototypes()
    
    # initialize the data index counter for the support examples
    # the client will use this counter to request the input data for this support example
    dataIndex = 0

    # get the string names of the labels that the model was trained on
    label_names = use_label_names(model, len(support_examples.keys()))

    # this list will hold all the prototype and support example embedding data for all labels
    embedding_data = []

    # loop over all labels
    for label_idx in support_examples.keys():
        prototype_and_support_data = {
            "label": label_names[label_idx] if label_names is not None else str(label_idx), # get the label name
            "prototype": prototypes[label_idx], # get the prototype
            "trainingExamples": [] # initialize a list for the support examples
        }

        # run inference on all the support examples to get the embedding data
        support_example_predictions = model.predict(support_examples[label_idx])
        
        # loop over all the support examples
        for support_idx in range(len(support_example_predictions.embeddings)):
            # append this support example embedding data to the trainingExamples list
            prototype_and_support_data["trainingExamples"].append({
                # get the embedding coordinates for this support example
                "coordinates": support_example_predictions.embeddings[support_idx],
                "inputData": { "dataIndex": dataIndex, },
                # get all the class confidence predictions for this support example
                "labels": [{
                    "label": label_names[nested_label_idx] if label_names is not None else str(nested_label_idx),
                    "confidence": d,
                } for nested_label_idx,d in enumerate(support_example_predictions.classes[support_idx])],
                # get the OOD score for this support example
                "ood": support_example_predictions.ood_scores[support_idx]
            })
            dataIndex += 1 # increment the data index counter

        embedding_data.append(prototype_and_support_data)

    return embedding_data

@convert_kwargs_to_snake_case
def resolve_dimensionality_reduction(_, info, method, data, n_neighbors, random_state=42):
    high_dimensions = len(data[0])
    num_samples = len(data)

    if method == "pca":
        # ideally we want to get up to 5 scree values
        # but that is not possible if the number of dimensions is less than 5
        technique = PCA(n_components=min(high_dimensions, 5)) # we have more than 2 components so we can get the scree values
    elif method == "tsne":
        technique = TSNE(n_components=2, perplexity=5, random_state=random_state)
    elif method == "mds":
        technique = MDS(n_components=2, normalized_stress=True, metric=False, random_state=random_state)
    else:
        technique = umap.UMAP(densmap=True, n_neighbors=min(high_dimensions, num_samples, n_neighbors), random_state=random_state, n_jobs=1) #densmap=True,
    data = np.array(data).astype(float)
    embeddings = technique.fit_transform(data)

    scree = None
    if isinstance(technique, PCA):
        scree = technique.explained_variance_ratio_.tolist()
        embeddings = embeddings[:,0:2] # slice off dimensions 3+ that we don't need

    zadu_obj = zadu.ZADU([
        { "id": "tnc", "params": {"k": n_neighbors} },
        { "id": "stress" },
        { "id": "srho" }
    ], data)
    scores = zadu_obj.measure(embeddings)

    trustworthiness = scores[0]["trustworthiness"]
    continuity = scores[0]["continuity"]
    stress = scores[1]["stress"]
    srho = scores[2]["spearman_rho"]

    return {
        "continuity": continuity,
        "embeddings": embeddings,
        "stress": stress,
        "scree": scree,
        "srho": srho,
        "trustworthiness": trustworthiness,
    }

@convert_kwargs_to_snake_case
def resolve_render_inference_feature_data(_, info, run_id, model_name, data_index):
    featureData, _, feature_names = get_sample_from_data_index(run_id, data_index, model_name=model_name)
    columnHeaders = feature_names if feature_names is not None else list(range(len(featureData)))
    assert len(featureData) == len(columnHeaders)
    
    return {"featureData": featureData, "columnHeaders": columnHeaders}

@convert_kwargs_to_snake_case
def resolve_render_support_feature_data(_, info, model_name, data_index):
    support_example, _, feature_names = get_support_example_from_data_index(model_name, data_index)
    featureData = support_example.tolist()
    columnHeaders = feature_names if feature_names is not None else list(range(len(featureData)))
    assert len(featureData) == len(columnHeaders)
    
    return {"featureData": featureData, "columnHeaders": columnHeaders}

def resolve_training_progress(_, info):
    pass

@convert_kwargs_to_snake_case
def resolve_download_model(_, info, model_name):
    pass



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import os

import time
import torch
import equine as eq
from ariadne import convert_kwargs_to_snake_case

from equine_webapp.utils import SERVER_CONFIG, combine_data_files, get_model_path, use_label_names

@convert_kwargs_to_snake_case
def resolve_upload_model(_, info, model_file):
    model_save_path = os.path.join(os.getcwd(), SERVER_CONFIG.MODEL_FOLDER_PATH, model_file.filename)
    model_file.save(model_save_path)

    if os.path.isfile(model_save_path):
        return {"success" : True}
    else:
        raise OSError("File not saved")
    
@convert_kwargs_to_snake_case
def resolve_upload_file(_, info, file): #TODO Deduplicate code from upload model endpoint
    save_path = os.path.join(os.getcwd(), SERVER_CONFIG.UPLOAD_FOLDER_PATH, file.name)
    file.save(save_path)

    if os.path.isfile(save_path):
        return {"success" : True}
    else:
        raise OSError("File not saved")

@convert_kwargs_to_snake_case
def resolve_run_inference(_, info, model_name, sample_filenames):
    run_id = int(time.time()) #TODO Better way to generate ID?
    
    # load the model
    model_path = get_model_path(model_name)
    model = eq.load_equine_model(model_path)
    input_dtype = next(model.embedding_model.parameters()).dtype
    
    # run inference on the samples
    sample_dataset = combine_data_files(sample_filenames)
    predictions = model.predict(sample_dataset.dataset.tensors[0].to(input_dtype))

    # get the string names of the labels that the model was trained on
    label_names = use_label_names(model, predictions.classes.shape[-1])
    
    # this list will hold all the samples data to send back to the client
    samples_json = []

    # loop through all the samples
    for sample_idx in range(len(sample_dataset.dataset)):
        json_data = {
            "coordinates": predictions.embeddings[sample_idx],
            "inputData": {
                "file": sample_dataset.filenames[sample_idx],
                "dataIndex" : sample_idx
            },
            "labels": [{
                "label": label_names[label_idx] if label_names is not None else str(label_idx),
                "confidence": d,
            } for label_idx,d in enumerate(predictions.classes[sample_idx])],
            "ood": predictions.ood_scores[sample_idx]
        }

        samples_json.append(json_data)

    save_filename = f"{run_id}_run_data.pt"
    torch.save(sample_dataset, os.path.join(SERVER_CONFIG.UPLOAD_FOLDER_PATH, save_filename))

    return {
        "samples": samples_json,
        "version": eq.__version__,
        "runId" : run_id
    }


@convert_kwargs_to_snake_case
def resolve_train_model(_, info, episodes, sample_filenames, embed_model_name, new_model_name, train_model_type, emb_out_dim = 0):
    model_file = embed_model_name if ".jit" in embed_model_name else embed_model_name + ".jit"
    model_path = os.path.join(os.getcwd(), SERVER_CONFIG.MODEL_FOLDER_PATH, model_file)
    embed_model = torch.jit.load(model_path)
    input_dtype = next(embed_model.parameters()).dtype

    sample_dataset = combine_data_files(sample_filenames, is_train=True)

    X = sample_dataset.dataset.tensors[0].to(input_dtype)
    Y = sample_dataset.dataset.tensors[1]

    dataset = torch.utils.data.TensorDataset(X, Y)
    num_classes = len(torch.unique(Y))
    if emb_out_dim == 0: emb_out_dim = num_classes

    if train_model_type == "EquineProtonet":
        model = eq.EquineProtonet(embed_model, emb_out_dim)
        model.train_model(dataset, num_episodes=episodes)
    elif train_model_type == "EquineGP":
        model = eq.EquineGP(embed_model, emb_out_dim, num_classes)

        loss_fn = torch.nn.CrossEntropyLoss()
        optimizer = torch.optim.SGD(
            model.parameters(),
            lr=0.001,
            momentum=0.9,
            weight_decay=0.0001,
        )
        model.train_model(dataset, num_epochs=episodes, loss_fn=loss_fn, opt=optimizer, vis_support=True, support_size=10)
    else:
        raise ValueError(f"Given train_model_type '{train_model_type}' is not valid.")
    
    model.save(os.path.join(os.getcwd(), SERVER_CONFIG.MODEL_FOLDER_PATH, new_model_name+SERVER_CONFIG.MODEL_EXT))

    return {"success": True}



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
from ariadne import QueryType, MutationType, make_executable_schema, load_schema_from_path
import equine_webapp.graphql.query_resolvers
import equine_webapp.graphql.mutation_resolvers
from equine_webapp.utils import SERVER_CONFIG

type_defs = load_schema_from_path(SERVER_CONFIG.SCHEMA_PATH)

query = QueryType()
query.set_field("models", equine_webapp.graphql.query_resolvers.resolve_available_models)
query.set_field("modelSummary", equine_webapp.graphql.query_resolvers.resolve_model_summary)
query.set_field("getPrototypeSupportEmbeddings", equine_webapp.graphql.query_resolvers.resolve_get_protonet_support_embeddings)
query.set_field("dimensionalityReduction", equine_webapp.graphql.query_resolvers.resolve_dimensionality_reduction)
query.set_field("renderInferenceFeatureData", equine_webapp.graphql.query_resolvers.resolve_render_inference_feature_data)
query.set_field("renderSupportFeatureData", equine_webapp.graphql.query_resolvers.resolve_render_support_feature_data)

mutation = MutationType()
mutation.set_field("uploadFile", equine_webapp.graphql.mutation_resolvers.resolve_upload_file)
mutation.set_field("uploadModel", equine_webapp.graphql.mutation_resolvers.resolve_upload_model)
mutation.set_field("runInference", equine_webapp.graphql.mutation_resolvers.resolve_run_inference)
mutation.set_field("startTraining", equine_webapp.graphql.mutation_resolvers.resolve_train_model)

schema = make_executable_schema(type_defs, query, mutation)


# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import numpy as np
import os
from sklearn.datasets import make_blobs
import torch

import equine as eq
from equine_webapp.utils import SERVER_CONFIG

TEST_MODEL_CONFIG = {
    "emb_out_dim": 16,
    "epochs": 50,
    "examples_per_class": 500,
    "model_name": "protonet_test_model.eq",
    "num_classes": 4,
    "protonet_num_episodes": 100,
    "support_size": 20,
    "tensor_dim": 2,
    "train_ratio": 0.8,
    "test_ratio": 0.2,
}

def train_model_for_testing(save_path):
    n_samples = TEST_MODEL_CONFIG["examples_per_class"] * TEST_MODEL_CONFIG["num_classes"]
    centers = [ (2,2), (-2,2), (-2,-2), (2,-2), (0,0) ]
    x_list, y_list = make_blobs(n_samples=n_samples, n_features=TEST_MODEL_CONFIG["tensor_dim"], 
                                centers=centers, cluster_std=0.75, shuffle=False,random_state=52)
    # Randomize the classes in the middle
    rng = np.random.default_rng(seed=52)
    y_list[(n_samples*4//5):] = rng.integers(4, size=n_samples//5) 
    X = torch.FloatTensor(x_list)
    Y = torch.tensor(y_list)

    dataset = torch.utils.data.TensorDataset(X,Y)
    trainset, testset = torch.utils.data.random_split(dataset, [TEST_MODEL_CONFIG["train_ratio"], TEST_MODEL_CONFIG["test_ratio"]], 
                        generator=torch.Generator().manual_seed(52))
    train_x = trainset.dataset.tensors[0][trainset.indices]
    train_y = trainset.dataset.tensors[1][trainset.indices]
    test_x  = testset.dataset.tensors[0][testset.indices]
    test_y  = testset.dataset.tensors[1][testset.indices]

    vanilla_nn = EmbeddingModel()
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(vanilla_nn.parameters())

    trainloader = torch.utils.data.DataLoader(trainset,batch_size=50,shuffle=True)

    vanilla_nn.train()
    for epoch in range(TEST_MODEL_CONFIG["epochs"]):
        epoch_loss = 0.0
        for i, (xs, labels) in enumerate(trainloader):
            optimizer.zero_grad()
            yhats = vanilla_nn(xs)
            loss = loss_fn(yhats, labels)
            loss.backward()
            optimizer.step()
            epoch_loss += loss.item()
        if epoch%50 == 49 or epoch == 0:
            print(f"Epoch {epoch+1} has loss {epoch_loss:3.2f}")

    # Load weights from the trained model
    sd = vanilla_nn.state_dict()
    em = EmbeddingModel()
    em.load_state_dict(sd)

    # Pull out the feature embedding
    all_layers = list(vanilla_nn.children())
    embedding_layers = all_layers[0][:-1] # Remove the last layer
    embedding_model = torch.nn.Sequential(*embedding_layers) # Rebuild the NN

    # Create the EQUINE model, support examples, and pointers to those examples 
    model = eq.EquineProtonet(embedding_model, emb_out_dim=TEST_MODEL_CONFIG["emb_out_dim"])

    model.train_model(torch.utils.data.TensorDataset(train_x,train_y),
                    way=4,             # Number of classes to train each episode
                    support_size=TEST_MODEL_CONFIG["support_size"],           # Number of support examples per class each episode
                    num_episodes=TEST_MODEL_CONFIG["protonet_num_episodes"], # Number of episodes (like epochs)
                    episode_size=100)  # Number training points selected per episode (like batches)
    model.save(save_path)

    test_df = pd.DataFrame( test_x.numpy() ) # convert the test data into a np array, then into a data frame
    test_df.to_csv(os.path.join(SERVER_CONFIG.UPLOAD_FOLDER_PATH, "test_no_labels.csv"),index=False) #save to a csv

    return

class EmbeddingModel(torch.nn.Module):
    def __init__(self):
        super(EmbeddingModel, self).__init__()
        self.linear_relu_stack = torch.nn.Sequential(
            torch.nn.Linear(TEST_MODEL_CONFIG["tensor_dim"], TEST_MODEL_CONFIG["emb_out_dim"]),
            torch.nn.ReLU(),
            torch.nn.Linear(TEST_MODEL_CONFIG["emb_out_dim"], TEST_MODEL_CONFIG["emb_out_dim"]),
            torch.nn.ReLU(),
            torch.nn.Linear(TEST_MODEL_CONFIG["emb_out_dim"], TEST_MODEL_CONFIG["emb_out_dim"]),
            torch.nn.ReLU(),
            torch.nn.Linear(TEST_MODEL_CONFIG["emb_out_dim"], TEST_MODEL_CONFIG["num_classes"])
        )
    def forward(self, x):
        logits = self.linear_relu_stack(x)
        return logits



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import time

from equine_webapp.tests.train_model_for_testing import TEST_MODEL_CONFIG


def test_query_models(client):
    response = client.post("/graphql", json={
        "query": """
            query QueryModels($extension:String!) {
              models(extension: $extension) {
                name
                lastModified
              }
            }
        """,
        "variables": {"extension": ".eq"},
    })
    results = response.json["data"]["models"]

    # this is the model we just trained in ariadne_server/tests/conftest.py
    assert results[0]["name"] == TEST_MODEL_CONFIG["model_name"]
    # check that the model was trained recently, since we won't know for sure when it will have finished
    assert abs(results[0]["lastModified"] - time.time()) < 100


# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import numpy as np
import os
import time

from equine_webapp.utils import SERVER_CONFIG
from equine_webapp.tests.utils import assert_confidence_labels_are_valid
from equine_webapp.tests.train_model_for_testing import TEST_MODEL_CONFIG

def test_mutation_runInference(client):
    response = client.post("/graphql", json={
        "query": """
            mutation Test($modelName: String!) {
              runInference(modelName: $modelName, sampleFilenames: ["test_no_labels.csv"]) {
                samples {
                  coordinates
                  inputData {
                    dataIndex
                  }
                  labels {
                    label
                    confidence
                  }
                  ood
                }
                runId
              }
            }
        """,
        "variables": {"modelName": TEST_MODEL_CONFIG["model_name"]},
    })


    # assert that the runInference run happened recently
    run_inference_result = response.json["data"]["runInference"]
    run_id = run_inference_result["runId"]
    assert abs(run_id - time.time()) < 100  

    # assert that the number of inference samples is correct
    test_data_size = int(TEST_MODEL_CONFIG["examples_per_class"]*TEST_MODEL_CONFIG["num_classes"]*TEST_MODEL_CONFIG["test_ratio"])
    assert len(run_inference_result["samples"]) == test_data_size

    # load the test data hard coded in train_model_for_testing.py
    test_df = pd.read_csv(os.path.join(SERVER_CONFIG.UPLOAD_FOLDER_PATH, "test_no_labels.csv"))

    # loop through all the samples to make sure they're correct
    for idx, sample in enumerate(run_inference_result["samples"]):
        # coordinates
        assert len(sample["coordinates"]) == TEST_MODEL_CONFIG["emb_out_dim"] # expected number of coorindates
        assert all(isinstance(item, float) for item in sample["coordinates"]) # all coordinates are floats

        # ood is between 0 and 1
        assert sample["ood"] >= 0.0
        assert sample["ood"] <= 1.0

        # input data
        data_index = sample["inputData"]["dataIndex"]
        assert data_index == idx

        # request the render data and assert that the right inference sample data is being sent
        render_inference_response = client.post("/graphql", json={
            "query": """
                query RenderInferenceFeatureData($runId: Int!, $modelName: String!, $dataIndex: Int!) {
                  renderInferenceFeatureData(runId:$runId, modelName: $modelName, dataIndex: $dataIndex){
                    featureData,
                    columnHeaders
                  }
                }
            """,
            "variables": {"dataIndex": data_index, "modelName": TEST_MODEL_CONFIG["model_name"], "runId": run_id},
        })
        render_inference_result = render_inference_response.json["data"]["renderInferenceFeatureData"]
        assert np.allclose(
            np.array(render_inference_result["featureData"]),
            np.array(test_df.iloc[idx].tolist()),
            atol=1e-6
        )
        assert render_inference_result["columnHeaders"] == ["0", "1"]
        
        # labels
        assert_confidence_labels_are_valid(sample["labels"])

# TODO test images?


# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import os

import torch
import equine as eq
from pathlib import Path
from typing import Optional

class Config:
    MODEL_EXT: str = ".eq"
    OUTPUT_FOLDER: str
    SCHEMA_PATH: str

    def __init__(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        self.SCHEMA_PATH = os.path.join(dir_path, "graphql/types")
        if os.environ.get('TESTING') == "True":
            self.OUTPUT_FOLDER = os.path.join(dir_path, "tests/temp")
        else:
            self.OUTPUT_FOLDER = os.path.join(dir_path, "webapp-output")
            
        self.MODEL_FOLDER_PATH = os.path.join(self.OUTPUT_FOLDER, "models/") #pylint: disable=no-member
        self.UPLOAD_FOLDER_PATH = os.path.join(self.OUTPUT_FOLDER, "uploads/") #pylint: disable=no-member
        Path(self.MODEL_FOLDER_PATH).mkdir(parents=True, exist_ok=True)
        Path(self.UPLOAD_FOLDER_PATH).mkdir(parents=True, exist_ok=True)

SERVER_CONFIG = Config()

class SampleDataset:
    def __init__(self, tensor_dataset, filenames, column_headers):
        self.dataset = tensor_dataset
        self.filenames = filenames
        self.column_headers = column_headers


def combine_data_files(filename_list, is_train=False):
    dataset_list = []
    dataset_filenames = []
    column_headers = []

    for filename in filename_list:
        file_path = os.path.join(os.getcwd(), SERVER_CONFIG.UPLOAD_FOLDER_PATH, filename)
        if not os.path.isfile(file_path):
            raise ValueError(f"Data File '{file_path}' not found")
        
        file_ext = os.path.splitext(filename)[1]
        if file_ext == ".csv":
            dataframe = pd.read_csv(file_path)
            if is_train:
                data_labels = dataframe["labels"].to_numpy()
                dataframe = dataframe.drop(["labels"], axis=1)
                torch_labels = torch.from_numpy(data_labels)
            
            data_array = dataframe.to_numpy()
            torch_data = torch.from_numpy(data_array)
            if is_train:
                tensor_dataset = torch.utils.data.TensorDataset(torch_data, torch_labels)
            else:
                tensor_dataset = torch.utils.data.TensorDataset(torch_data)

            column_headers = dataframe.columns
            
        elif file_ext == ".pt":
            tensor_dataset = torch.load(file_path)
        else:
            raise ValueError(f"Given file '{filename} has unsupported file type '{file_ext}'")
        
        dataset_filenames += [filename]*len(tensor_dataset.tensors[0])
        dataset_list.append(tensor_dataset)

    dataset = torch.concat([data.tensors[0] for data in dataset_list], dim=0)
    
    if is_train:
        dataset_labels = torch.concat([data.tensors[1] for data in dataset_list], dim=0)
        tensor_dataset = torch.utils.data.TensorDataset(dataset, dataset_labels) #TODO Typechecking?
    else:
        tensor_dataset = torch.utils.data.TensorDataset(dataset) #TODO Typechecking?

    if len(column_headers) == 0:
        column_headers = [x for x in range(0, tensor_dataset.tensors[0].shape[1])]
    
    return SampleDataset(tensor_dataset, dataset_filenames, column_headers)


def get_support_example_from_data_index(model_name, data_index):
    data_index = int(data_index)
    assert data_index >= 0
    
    model_path = get_model_path(model_name)
    model = eq.load_equine_model(model_path)
    support = model.get_support()
    feature_names = model.get_feature_names()

    num_classes = len(support)
    num_support_per_class = support[0].shape[0]
    assert num_classes*num_support_per_class > data_index

    support_idx = data_index % num_support_per_class
    class_idx = int(data_index / num_support_per_class)

    return support[class_idx][support_idx], support, feature_names

def get_sample_from_data_index(run_id, data_index, model_name:Optional[str]=None):
    data_index = int(data_index)
    assert data_index >= 0

    feature_names = None
    if model_name is not None:
        model_path = get_model_path(model_name)
        model = eq.load_equine_model(model_path)
        feature_names = model.get_feature_names()

    run_id = int(run_id)
    sample_dataset = torch.load(os.path.join(SERVER_CONFIG.UPLOAD_FOLDER_PATH, f"{run_id}_run_data.pt"))
    assert len(sample_dataset.dataset) > data_index
    
    return sample_dataset.dataset[data_index][0], sample_dataset, feature_names

def get_model_path(model_name:str):
    model_file = model_name if SERVER_CONFIG.MODEL_EXT in model_name else model_name + SERVER_CONFIG.MODEL_EXT
    base_path = os.getcwd()
    model_path = sanitize_path(os.path.join(base_path, SERVER_CONFIG.MODEL_FOLDER_PATH, model_file), base_path)
    if not os.path.isfile(model_path):
        raise ValueError(f"Model File '{model_path}' not found")
    return model_path

def sanitize_path(file_path, base_path):
    fullpath = os.path.normpath(file_path)
    if not fullpath.startswith(base_path):
        raise Exception("Path not allowed")
    return fullpath

def use_label_names(model, num_labels:int):
    label_names = model.get_label_names()
    if label_names is not None and len(label_names) != num_labels:
        print(f"The number of label names ({len(label_names)}) does not match the number of classes ({num_labels}) in the predictions. The server will ignore the label names.")
        return None
    return label_names



# This script is used to evaluate the dimensionality reduction (DR) metrics for ScatterUQ
# You have to provide the model_files and test_sample_files
# This script does DR in the same way as ScatterUQ, evaluates the DR metrics,
# and prints the average metrics and standard deviations
# from equine import load_equine_model

# there's a weird bug when we need to import torch first before importing zadu (which imports faiss). if you don't import torch first, the faiss import in zadu will cause a seg fault 11 when you try to call torch.linalg.qr
import torch

import heapq
import numpy as np

from src.equine_webapp.graphql.query_resolvers import resolve_get_protonet_support_embeddings, resolve_dimensionality_reduction
from src.equine_webapp.graphql.mutation_resolvers import resolve_run_inference

# limit the number of samples we get per condition to avoid skewing our metrics too much
NUM_SAMPLES_PER_CONDITION=10

# these are the model files we want to evaluate
model_files = [
    "vis_MNIST_GP_32D.eq",
    "vis_MNIST_GP_1024D.eq",
]

# these are the files of test samples to evaluate on
test_sample_files = [
    "some_digits.pt",
    "some_fashion.pt",
]


def get_metrics_for_method(samples, uq_viz_data, method, outlier_tolerance, class_confidence_threshold):
    # this diciontary tracks the DR metrics for each condition
    metrics_dict = {
        "ood": [],
        "confident": [],
        "confused_class": [],
    }

    for sample in samples: # loop through all the samples
        # determine which condition this sample falls under
        condition = "" # refers to whether the sample is in the OOD, confident, or class confusion use case
        if sample["ood"] > outlier_tolerance: # if this is an OOD sample
            condition = "ood"
        elif any(l["confidence"]>class_confidence_threshold for l in sample["labels"]): # if this is a confident sample
            condition = "confident"
        else: # else the model is confused about which class
            condition = "confused_class"

        # if we have too many samples in this condition, skip this sample
        if len(metrics_dict[condition]) >= NUM_SAMPLES_PER_CONDITION:
            continue
        
        dr_points = [sample["coordinates"]] # add this sample to the points for DR
        if condition=="ood" or condition=="confident": # if this is an OOD or confident sample
            # the DR in ScatterUQ is the same for OOD and confident conditions
            # we need to find the closest, ie most confident, label + training examples for DR
            # "training examples" are synonymous with "support examples"
            closest_label = max(
                sample["labels"], 
                key=lambda l:l["confidence"]
            )
            closest_label = next( # get the coordinates for the highest label
                x for x in uq_viz_data if x["label"]==closest_label["label"]
            )
            dr_points = dr_points + [closest_label["prototype"]] # add the prototype
            dr_points = dr_points + [t["coordinates"] for t in closest_label["trainingExamples"]] # add the training examples
        else: # else the model is confused about which class the sample belongs to
            # we need to find the two closest labels + training examples for DR 
            [most_confident_label, second_confident_label] = heapq.nlargest(
                2, sample["labels"],
                key=lambda l:l["confidence"]
            )
            most_confident_label = next( # get the coordinates
                x for x in uq_viz_data if x["label"]==most_confident_label["label"]
            )
            second_confident_label = next( # get the coordinates
                x for x in uq_viz_data if x["label"]==second_confident_label["label"]
            )
            if(most_confident_label["label"] == second_confident_label["label"]):
                raise ValueError("The most and second most confident labels should not be the same")
            
            dr_points = dr_points + [most_confident_label["prototype"]] # add the closest prototype
            dr_points = dr_points + [t["coordinates"] for t in most_confident_label["trainingExamples"]] # add the training examples
            dr_points = dr_points + [second_confident_label["prototype"]] # add the second closest prototype
            dr_points = dr_points + [t["coordinates"] for t in second_confident_label["trainingExamples"]] # add the training examples
        
        dr_points = list(p.detach().numpy() for p in dr_points) # detach the pytorch tensors for numpy
        # do DR and get the metrics
        metrics_dict[condition].append(
            resolve_dimensionality_reduction(None, {}, method=method, data=dr_points, n_neighbors=5)
        )

    # print("ood count",len(metrics_dict["ood"]),"confused_class count",len(metrics_dict["confused_class"]),"confident count",len(metrics_dict["confident"]))

    return metrics_dict



def get_metrics_as_np_arrays(metrics_list, title, method):
    # time = np.array([m['time'] for m in metrics_list])
    trust = np.array([m['trustworthiness'] for m in metrics_list])
    continuity = np.array([m['continuity'] for m in metrics_list])
    stress = np.array([m['stress'] for m in metrics_list])
    srho = np.array([m['srho'] for m in metrics_list])

    print(f"Method: {method} {title}, Continuity: {format_avg_std(continuity)},  Stress: { format_avg_std(stress)}, Spearman’s R: {format_avg_std(srho)},  Trust: {format_avg_std(trust)}")

    return continuity, srho, stress, trust

def format_avg_std(np_array):
    return f"{np.average(np_array):.10f} (± {np.std(np_array):.10f})"



if __name__ == "__main__":
    outlier_tolerances = [0.2,0.4,0.6,0.8]
    class_confidence_thresholds = [0.2,0.4,0.6,0.8]

    # iterate over the models we are evaluating
    for model_file in model_files:
        metrics_list = [] # this list will hold all the metrics for this model

        # run inference on all the test samples from the files
        inference_data = resolve_run_inference(None, {}, model_file, test_sample_files)
        samples = inference_data["samples"]

        # get the prototype and support example embeddings for this model
        uq_viz_data = resolve_get_protonet_support_embeddings(None, {}, model_file)

        # sweep over several combinations of outlier tolerances and class confidence thresholds
        for outlier_tolerance in outlier_tolerances:
            for class_confidence_threshold in class_confidence_thresholds:
                # calculate metrics
                metrics_dict = get_metrics_for_method(
                    samples=samples, 
                    uq_viz_data=uq_viz_data, 
                    method="pca",
                    outlier_tolerance=outlier_tolerance,
                    class_confidence_threshold=class_confidence_threshold
                )

                # append all the metrics into the main metrics list
                metrics_list = metrics_list + metrics_dict["ood"] + metrics_dict["confident"] + metrics_dict["confused_class"]

        get_metrics_as_np_arrays(metrics_list, model_file, "pca")

    


# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import math

from equine_webapp.tests.train_model_for_testing import TEST_MODEL_CONFIG

def assert_confidence_labels_are_valid(labels):
    assert len(labels) == TEST_MODEL_CONFIG["num_classes"] # correct number of classes
    sum_confidences = 0.0
    for label_idx, label in enumerate(labels):
        assert str(label_idx) == label["label"]

        # the confidence is between 0 and 1
        assert label["confidence"] >= 0.0
        assert label["confidence"] <= 1.0
        sum_confidences += label["confidence"] # sum the confidence scores
    assert math.isclose(sum_confidences, 1.0, abs_tol=0.00001) # the confidences sum close to 1


# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

def test_static_web_files(client):
    response = client.get("/")
    is_html_str(response.text)
    assert "<title>EQUINE Webapp</title>" in response.text


    response = client.get("/dashboard")
    is_html_str(response.text)
    assert "<title>EQUINE Webapp</title>" in response.text

    response = client.get("/demo")
    is_html_str(response.text)
    assert "<title>EQUINE Webapp</title>" in response.text

    response = client.get("/download")
    is_html_str(response.text)
    assert "<title>EQUINE Webapp</title>" in response.text

    response = client.get("/model-summary-page")
    is_html_str(response.text)
    assert "<title>EQUINE Webapp</title>" in response.text

    response = client.get("/settings")
    is_html_str(response.text)
    assert "<title>EQUINE Webapp</title>" in response.text

    response = client.get("/training")
    is_html_str(response.text)
    assert "<title>EQUINE Webapp</title>" in response.text


def is_html_str(text:str):
    assert text.startswith('<!DOCTYPE html><html lang="en">')
    assert text.endswith('</html>')


# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import os
import pytest
import shutil

os.environ["TESTING"] = "True"

from equine_webapp.flask_server import app as my_app
from equine_webapp.utils import SERVER_CONFIG
from equine_webapp.tests.train_model_for_testing import train_model_for_testing, TEST_MODEL_CONFIG

@pytest.fixture()
def app():
    my_app.config.update({ "TESTING": True })
    # other setup can go here
    yield my_app
    # clean up / reset resources here


@pytest.fixture()
def client(app):
    return app.test_client()


@pytest.fixture()
def runner(app):
    return app.test_cli_runner()


def pytest_configure():
    train_model_for_testing(os.path.join(SERVER_CONFIG.MODEL_FOLDER_PATH, TEST_MODEL_CONFIG["model_name"]))


def pytest_sessionfinish(session, exitstatus):
    shutil.rmtree(SERVER_CONFIG.OUTPUT_FOLDER)
    pass



# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT
import equine as eq
import numpy as np
import os

from equine_webapp.tests.train_model_for_testing import TEST_MODEL_CONFIG
from equine_webapp.tests.utils import assert_confidence_labels_are_valid
from equine_webapp.utils import SERVER_CONFIG

def test_query_models(client):
    model_path = os.path.join(os.getcwd(), SERVER_CONFIG.MODEL_FOLDER_PATH, "protonet_test_model.eq")
    model = eq.load_equine_model(model_path)
    model_support = model.model.support



    response = client.post("/graphql", json={
        "query": """
            query GetPrototypeSupportEmbeddings($modelName:String!) {
              getPrototypeSupportEmbeddings(modelName:$modelName) {
                label
                prototype
                trainingExamples {
                  coordinates
                  inputData {
                    dataIndex
                  }
                  labels {
                    label
                    confidence
                  }
                  ood
                }
              }
            }
        """,
        "variables": {"modelName": TEST_MODEL_CONFIG["model_name"]},
    })

    embeddings = response.json["data"]["getPrototypeSupportEmbeddings"]
    assert len(embeddings) == TEST_MODEL_CONFIG["num_classes"] # correct number of classes
    for label_idx, embedding in enumerate(embeddings):
        assert str(label_idx) == embedding["label"] # correct label

        # embeddings
        assert len(embedding["prototype"]) == TEST_MODEL_CONFIG["emb_out_dim"] # correct embedding vector size
        all(isinstance(item, float) for item in embedding["prototype"]) # all coordinates are floats

        for support_idx, support in enumerate(embedding["trainingExamples"]):
            # coordinates
            assert len(support["coordinates"]) == TEST_MODEL_CONFIG["emb_out_dim"] # correct embedding vector size
            assert all(isinstance(item, float) for item in support["coordinates"]) # all coordinates are floats

            # ood is between 0 and 1
            assert support["ood"] >= 0.0
            assert support["ood"] <= 1.0

            # input data
            data_index = support["inputData"]["dataIndex"]
            support_data_idx = TEST_MODEL_CONFIG["support_size"]*label_idx + support_idx
            assert data_index == support_data_idx
            
            # request the render data and assert that the right support example data is being sent
            render_support_response = client.post("/graphql", json={
                "query": """
                    query RenderSupportFeatureData($modelName: String!, $dataIndex: Int!) {
                      renderSupportFeatureData(modelName:$modelName, dataIndex: $dataIndex){
                        featureData,
                        columnHeaders
                      }
                    }
                """,
                "variables": {"dataIndex": data_index, "modelName": TEST_MODEL_CONFIG["model_name"]},
            })
            render_support_result = render_support_response.json["data"]["renderSupportFeatureData"]
            assert np.allclose(
                np.array(render_support_result["featureData"]),
                np.array(model_support[label_idx][support_idx]),
                atol=1e-6
            )
            assert render_support_result["columnHeaders"] == ["0", "1"]

            # labels
            assert_confidence_labels_are_valid(support["labels"])

# TODO test images?


# Copyright (c) 2023 Massachusetts Institute of Technology
# SPDX-License-Identifier: MIT

import json
import numpy as np
import os
from pathlib import Path

from equine_webapp.utils import SERVER_CONFIG


def test_mutation_uploadFile(client):
    # make a temporary test folder
    folder = os.path.join(SERVER_CONFIG.OUTPUT_FOLDER, "upload_model_test/")
    Path(folder).mkdir(parents=True, exist_ok=True)
    
    # save a dummy csv file
    test_df = pd.DataFrame( [[1.0,2.0,3.0],[4.0,5.0,6.0]] )
    test_df.to_csv(os.path.join(folder, "test_upload_data_file.csv"),index=False, index_label=False)

    #https://flask.palletsprojects.com/en/2.3.x/testing/#form-data
    response = client.post("/graphql", data={
        "map": json.dumps({"test_upload_data_file.csv": ["variables.file"]}),
        "test_upload_data_file.csv": (Path(folder) / "test_upload_data_file.csv").open("rb"),
        "operations": json.dumps({
            "query": """mutation UploadSampleFile($file: Upload!) {
              uploadFile(file:$file) {
                success
              }
            }""",
            "variables": { "file": None}
        }),
    })
    
    assert response.json["data"]["uploadFile"]["success"]

    uploaded_df = pd.read_csv(
        os.path.join(SERVER_CONFIG.UPLOAD_FOLDER_PATH, "test_upload_data_file.csv"),
    )
    assert np.array_equal(test_df.values,uploaded_df.values)